"""
Influence Types and Use Case Examples

1. Sorts/Searches: Define choice set sortings under specific conditions
2. Defaults: Set order field value defaults under specific conditions
3. Recommendations: Suggest documentation, orders, analyses, and patient engagements
4. Rationales: Request or require rationales when behavior diverges from protocols
5. Tasks: Assign a task to a user

Below is a small sample of the 200+ events in charting and coordination available for subscription.

Charting Events

CHART_OPENED, CHART_QUERIED, PRESCRIPTION_ORIGINATED, PRESCRIPTION_COMMITTED, REFERRAL_ORIGINATED,
REFERRAL_COMMITTED, NOTE_LOCKED, ...

Care Coordination Events

APPOINTMENT_SCHEDULED, PATIENT_ARRIVED, PATIENT_ROOMED, REFILL_REQUEST_RECEIVED,
REFERRAL_LOOP_CLOSED, POPULATION_QUERIED, MULTIPATIENT_ACTION_ORIGINATED, ...
"""

import importlib
import pkgutil

from collections import defaultdict

import arrow

from .patient import Patient
from .timeframe import Timeframe


class Protocol(object):

    responds_to_event_types = []

    def __init__(self, patient, **kwargs):
        self.patient = patient


class ClinicalQualityMeasure(Protocol):
    """
    Example CQMs:

    - pneumonia
    - depression screening
    """

    def __init__(self, timeframe=None, **kwargs):
        super().__init__(**kwargs)

        if timeframe:
            self.timeframe = timeframe
        else:
            now = arrow.utcnow()

            self.timeframe = Timeframe(start=now, end=now)

    def get_narrative_recommendations(self):
        return {}

    def in_initial_population(self):
        raise NotImplementedError('in_initial_population must be overridden')

    def in_denominator(self):
        raise NotImplementedError('in_denominator must be overridden')

    def in_numerator(self):
        raise NotImplementedError('in_numerator must be overridden')

    # TODO move to ClinicalQualityMeasureReport class
    # def report(self):
    #     """
    #     Report aggregate statistical information on patients who are included or excluded from this
    #     protocol's criteria.
    #     """
    #     return {}


def get_subclasses(cls):
    subclasses = []

    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(get_subclasses(subclass))

    return subclasses


def get_protocols(module):
    for _, name, _ in pkgutil.iter_modules(module.__path__):
        importlib.import_module(f'{module.__name__}.{name}')

    subclasses = get_subclasses(Protocol)
    protocols_by_event_type = defaultdict(list)

    for subclass in subclasses:
        if subclass in (Protocol, ClinicalQualityMeasure):
            continue

        for event_type in subclass.responds_to_event_types:
            protocols_by_event_type[event_type].append(subclass)

    return protocols_by_event_type


def protocols_for_patient(protocols, event_type, patient_key):
    patient = Patient(patient_key)

    return [protocol(patient=patient) for protocol in protocols[event_type]]



import arrow

from . import cds_client

CODE_SYSTEM_MAPPING = {
    'icd10cm': 'ICD-10',
    'icd10pcs': 'ICD-10',
    'snomedct': 'http://snomed.info/sct',
}

SYSTEM_CODE_MAPPING = {
    'ICD-10': ['icd10cm', 'icd10pcs'],
    'http://snomed.info/sct': ['snomedct'],
}

VALID_CONDITION_SYSTEMS = [
    'icd10cm',
    'icd10pcs',
    'snomedct',
]

VALID_LAB_SYSTEMS = [
    'loinc',
]


def coalesce(value_set, kwargs):
    if value_set:
        return value_set.values

    return kwargs


class Patient(object):
    """
    Wrapper around a Canvas patient to add convenience methods specific to science functionality.
    """

    def __init__(self, patient_key):
        # TODO replace this with call to CDS 'Chart' DRF view (to be created) that serializes
        # labs/meds/everything else
        self.patient = cds_client.get_patient(patient_key)

        self.conditions = cds_client.get_patient_conditions(patient_key)
        self.lab_results = cds_client.get_patient_conditions(patient_key)

    @property
    def first_name(self):
        return self.patient['firstName']

    @property
    def age(self):
        return float(self.patient['age'])

    # TODO timezones
    def age_at(self, time):
        birth_date = arrow.get(self.patient['birthDate'])

        return (time - birth_date).years

    def vaccines_within(self, timeframe, value_set=None, **kwargs):
        pass

    #   ____                _ _ _   _
    #  / ___|___  _ __   __| (_) |_(_) ___  _ __  ___
    # | |   / _ \| '_ \ / _` | | __| |/ _ \| '_ \/ __|
    # | |__| (_) | | | | (_| | | |_| | (_) | | | \__ \
    #  \____\___/|_| |_|\__,_|_|\__|_|\___/|_| |_|___/
    #

    def gather_conditions(self, **kwargs):
        conditions = []
        systems = [key for key, _ in kwargs.items() if key in VALID_CONDITION_SYSTEMS]

        for condition in self.conditions:
            if 'code' not in condition['resource']:
                continue

            for coding in condition['resource']['code']['coding']:
                coding_systems = SYSTEM_CODE_MAPPING[coding['system']]

                for coding_system in coding_systems:
                    if coding_system in systems:
                        if coding['code'] in kwargs[coding_system]:
                            conditions.append(coding)

        return conditions

    def filter_conditions(self, filter_fn, **kwargs):
        conditions = self.gather_conditions(**kwargs)

        return [condition for condition in conditions
                if condition['onsetDate'] and
                filter_fn(arrow.get(condition['onsetDate']))]

    def conditions_within(self, timeframe, value_set=None, **kwargs):
        """
        Called with e.g. kwargs = {'loinc': [...]} or {'icd10cm': [...]}
        """
        def filter_fn(onset):
            return onset >= timeframe.start and onset <= timeframe.end

        return self.filter_conditions(filter_fn, **coalesce(value_set, kwargs))

    def conditions_before(self, end, value_set=None, **kwargs):
        return self.filter_conditions(lambda onset: onset <= end, **coalesce(value_set, kwargs))

    def conditions_after(self, start, value_set=None, **kwargs):
        return self.filter_conditions(lambda onset: onset >= start, **coalesce(value_set, kwargs))

    # TODO move to Protocol class so we have access to self.timeframe; cleaner
    def conditions_after_one_year_before_timeframe_end(self, timeframe, value_set=None, **kwargs):
        return self.conditions_within(timeframe.end.shift(years=-1), **coalesce(value_set, kwargs))

    #  _          _
    # | |    __ _| |__  ___
    # | |   / _` | '_ \/ __|
    # | |__| (_| | |_) \__ \
    # |_____\__,_|_.__/|___/
    #

    def gather_lab_results(self, loinc):
        lab_results = []

        for lab_result in self.lab_results:
            for value in lab_result['values']:
                if value['loincNum'] in loinc:
                    lab_results.append(value)

        return lab_results

    def filter_lab_results(self, filter_fn, **kwargs):
        lab_results = self.gather_lab_results(**kwargs)

        return [lab_result for lab_result in lab_results
                if filter_fn(arrow.get(lab_result['created']))]

    def lab_results_within(self, timeframe, value_set=None, **kwargs):
        """
        Call with e.g. kwargs = {'loinc': [...]} or {'icd10cm': [...]}
        """
        def filter_fn(effective):
            return effective >= timeframe.start and effective <= timeframe.end

        return self.filter_lab_results(filter_fn, **coalesce(value_set, kwargs))

    def most_recent_result_within(self, timeframe, value_set=None, **kwargs):
        # TODO this will change from 'created' to 'effective' once that field exists on CDS
        lab_results = sorted(
            self.lab_results_within(timeframe, **coalesce(value_set, kwargs)),
            key=lambda lab_result: arrow.get(lab_result['created']),
            reverse=True)

        return lab_results[0]



class ValueSystems(type):

    @property
    def values(cls):
        return {system.lower(): getattr(cls, system)
                for system in cls.value_systems
                if hasattr(cls, system)}


class ValueSet(object, metaclass=ValueSystems):

    value_systems = [
        'CPT',
        'CVX',
        'HCPCS',
        'ICD10CM',
        'ICD10PCS',
        'ICD9CM',
        'LOINC',
        'RXNORM',
        'SNOMEDCT',
    ]



import os

from requests import Session

CLINICAL_DATA_URL = os.getenv('CLINICAL_DATA_URL', 'https://cab.canvasmedical.com')
PRE_SHARED_KEY = os.getenv('PRE_SHARED_KEY')

home_client = Session()
home_client.headers.update({'Authorization': PRE_SHARED_KEY})


def paginate(url, patient_key, page_size=100):
    results = []
    offset = 0
    total = -1

    # XXX could run indefinitely if condition is added after first loop, would cursor pagination be
    # better?
    while len(results) != total:
        response = home_client.get(f'{CLINICAL_DATA_URL}{url}', params={
            'patient__key': patient_key,
            'offset': offset,
            'limit': page_size,
        })

        response_json = response.json()

        if total < 0:
            total = response_json['total']

        results.extend(response_json['entry'])

        offset += page_size

    return results


def get_patient(patient_key):
    response = home_client.get(f'{CLINICAL_DATA_URL}/api/Patient/{patient_key}')

    if response.status_code != 200:
        raise Exception(f"Patient '{patient_key}' not found")

    return response.json()


def get_patient_conditions(patient_key):
    return paginate('/api/Condition/', patient_key=patient_key)


def get_patient_lab_orders(patient_key):
    return paginate('/api/LabOrder/', patient_key=patient_key)


def get_patient_lab_results(patient_key):
    return paginate('/api/LabResult/', patient_key=patient_key)


def get_patient_procedures(patient_key):
    return paginate('/api/Procedure/', patient_key=patient_key)


def get_patient_encounters(patient_key):
    return paginate('/api/Encounter/', patient_key=patient_key)



class Timeframe(object):

    def __init__(self, start, end):
        self.start = start
        self.end = end

    def within(self, time):
        return time >= self.start and time <= self.end

    def add(self, time):
        pass

    def subtract(self, time):
        pass



class Recommendation(object):

    def __init__(self, *args, title, status_display, narrative, **kwargs):
        self.title = title
        self.status_display = status_display
        self.narrative = narrative

    def to_json(self):
        pass


class VaccineRecommendation(Recommendation):

    pass



from setuptools import find_packages, setup

setup(
    name='canvas-science',
    version='4.0.0',
    author='Canvas Medical',
    author_email='engineering@canvasmedical.com',
    url='https://github.com/canvas-medical/science-sdk',
    description="Canvas Medical's Science SDK",
    setup_requires=['setuptools-markdown'],
    long_description_markdown_filename='README.md',
    keywords='healthcare science sdk ehr',
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Environment :: Console',
        'Environment :: Web Environment',

        'Intended Audience :: Developers',
        'Intended Audience :: Healthcare Industry',

        'Operating System :: MacOS :: MacOS X',
        'Operating System :: POSIX',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',

        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(),
    install_requires=[
        'arrow>=0.10',
        'requests>=2.18',
    ])
